@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2020-2021 Fanout, Inc.
+ * Copyright (C) 2020-2022 Fanout, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -14,15 +14,14 @@
  * limitations under the License.
  */
 
-use crate::buffer::{write_vectored_offset, LimitBufs, VECTORED_MAX};
+use crate::buffer::{write_vectored_offset, write_vectored_offset_async, LimitBufs, VECTORED_MAX};
+use crate::future::AsyncWrite;
 use std::cmp;
 use std::convert::TryFrom;
 use std::io;
 use std::io::{Read, Write};
 use std::str;
 
-pub const HEADERS_MAX: usize = 32;
-
 const CHUNK_SIZE_MAX: usize = 0xffff;
 const CHUNK_HEADER_SIZE_MAX: usize = 6; // ffff\r\n
 const CHUNK_FOOTER: &[u8] = b"\r\n";
@@ -165,6 +164,82 @@ fn write_chunk(
     Ok(data_size)
 }
 
+// writes src to dest as chunks. current chunk state is passed in
+async fn write_chunk_async(
+    content: &[&[u8]],
+    footer: &[u8],
+    dest: &mut dyn AsyncWrite,
+    chunk: &mut Option<Chunk>,
+    max_size: usize,
+) -> Result<usize, io::Error> {
+    assert!(max_size <= CHUNK_SIZE_MAX);
+
+    let mut content_len = 0;
+    for buf in content.iter() {
+        content_len += buf.len();
+    }
+
+    if chunk.is_none() {
+        let size = cmp::min(content_len, max_size);
+
+        let mut h = [0; CHUNK_HEADER_SIZE_MAX];
+
+        let h_len = {
+            let mut c = io::Cursor::new(&mut h[..]);
+            write!(&mut c, "{:x}\r\n", size).unwrap();
+
+            c.position() as usize
+        };
+
+        *chunk = Some(Chunk {
+            header: h,
+            header_len: h_len,
+            size,
+            sent: 0,
+        });
+    }
+
+    let chunkv = chunk.as_mut().unwrap();
+
+    let cheader = &chunkv.header[..chunkv.header_len];
+    let data_size = chunkv.size;
+
+    let total = cheader.len() + data_size + footer.len();
+
+    let mut content_arr = [&b""[..]; VECTORED_MAX - 2];
+    for (i, buf) in content.iter().enumerate() {
+        content_arr[i] = buf;
+    }
+
+    let trim_content = (&mut content_arr[..content.len()]).limit(data_size);
+
+    let mut out_arr = [&b""[..]; VECTORED_MAX];
+    let mut out_arr_len = 0;
+
+    out_arr[0] = cheader;
+    out_arr_len += 1;
+
+    for buf in trim_content.iter() {
+        out_arr[out_arr_len] = buf;
+        out_arr_len += 1;
+    }
+
+    out_arr[out_arr_len] = footer;
+    out_arr_len += 1;
+
+    let size = write_vectored_offset_async(dest, &out_arr[..out_arr_len], chunkv.sent).await?;
+
+    chunkv.sent += size;
+
+    if chunkv.sent < total {
+        return Ok(0);
+    }
+
+    *chunk = None;
+
+    Ok(data_size)
+}
+
 #[cfg(test)]
 pub fn write_headers(writer: &mut dyn io::Write, headers: &[Header]) -> Result<(), io::Error> {
     for h in headers.iter() {
@@ -649,6 +724,80 @@ impl<'buf, 'headers> ServerProtocol {
 
         Ok(content_written)
     }
+
+    pub async fn send_body_async(
+        &mut self,
+        writer: &mut dyn AsyncWrite,
+        src: &[&[u8]],
+        end: bool,
+        headers: Option<&[u8]>,
+    ) -> Result<usize, ServerError> {
+        assert_eq!(self.state, ServerState::SendingBody);
+
+        let mut src_len = 0;
+        for buf in src.iter() {
+            src_len += buf.len();
+        }
+
+        if let BodySize::NoBody = self.body_size {
+            // ignore the data
+
+            if end {
+                self.state = ServerState::Finished;
+            }
+
+            return Ok(src_len);
+        }
+
+        if !self.chunked {
+            let size = write_vectored_offset_async(writer, src, 0).await?;
+
+            if end && size >= src_len {
+                self.state = ServerState::Finished;
+            }
+
+            return Ok(size);
+        }
+
+        // chunked
+
+        let mut content_written = 0;
+
+        if src_len > 0 {
+            content_written = write_chunk_async(
+                src,
+                CHUNK_FOOTER,
+                writer,
+                &mut self.sending_chunk,
+                CHUNK_SIZE_MAX,
+            )
+            .await?;
+        }
+
+        // if all content is written then we can send the closing chunk
+        if end && content_written >= src_len {
+            let footer = if let Some(headers) = headers {
+                headers
+            } else {
+                CHUNK_FOOTER
+            };
+
+            write_chunk_async(
+                &[b""],
+                footer,
+                writer,
+                &mut self.sending_chunk,
+                CHUNK_SIZE_MAX,
+            )
+            .await?;
+
+            if self.sending_chunk.is_none() {
+                self.state = ServerState::Finished;
+            }
+        }
+
+        Ok(content_written)
+    }
 }
 
 #[cfg(test)]
@@ -656,6 +805,8 @@ mod tests {
     use super::*;
     use std::mem;
 
+    const HEADERS_MAX: usize = 32;
+
     struct MyBuffer {
         data: Vec<u8>,
         max: usize,
@@ -858,12 +1009,36 @@ mod tests {
 
             body_out.max += write_size;
 
-            let size = match p.send_body(&mut body_out, &[&resp.body[sent..]], true, None) {
-                Ok(size) => size,
-                Err(ServerError::Io(e)) if e.kind() == io::ErrorKind::WriteZero => 0,
-                Err(_) => panic!("send_body failed"),
+            let trailing_headers = if !resp.trailing_headers.is_empty() {
+                let mut buf = Vec::new();
+
+                for (name, value) in resp.trailing_headers.iter() {
+                    write!(buf, "{}: ", name).unwrap();
+                    buf.write(value).unwrap();
+                    write!(buf, "\r\n").unwrap();
+                }
+
+                write!(buf, "\r\n").unwrap();
+
+                Some(buf)
+            } else {
+                None
             };
 
+            let trailing_headers: Option<&[u8]> = if let Some(trailing_headers) = &trailing_headers
+            {
+                Some(trailing_headers)
+            } else {
+                None
+            };
+
+            let size =
+                match p.send_body(&mut body_out, &[&resp.body[sent..]], true, trailing_headers) {
+                    Ok(size) => size,
+                    Err(ServerError::Io(e)) if e.kind() == io::ErrorKind::WriteZero => 0,
+                    Err(_) => panic!("send_body failed"),
+                };
+
             sent += size;
         }
 
@@ -1733,398 +1908,398 @@ mod tests {
         }
 
         let tests = [
-			Test {
-				name: "cant-write-1.1",
-				write_space: 5,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::Known(0),
-				ver_min: 1,
-				persistent: false,
-				result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
-				state: ServerState::AwaitingResponse,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "",
-			},
-			Test {
-				name: "cant-write-1.0",
-				write_space: 5,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::Known(0),
-				ver_min: 0,
-				persistent: false,
-				result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
-				state: ServerState::AwaitingResponse,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "",
-			},
-			Test {
-				name: "cant-write-status-line",
-				write_space: 12,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::Known(0),
-				ver_min: 0,
-				persistent: false,
-				result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
-				state: ServerState::AwaitingResponse,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.0 200",
-			},
-			Test {
-				name: "cant-write-header-name",
-				write_space: 20,
-				code: 200,
-				reason: "OK",
-				headers: &[
-					Header { name: "Foo", value: b"Bar" },
-				],
-				body_size: BodySize::Known(0),
-				ver_min: 0,
-				persistent: false,
-				result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
-				state: ServerState::AwaitingResponse,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.0 200 OK\r\nFoo",
-			},
-			Test {
-				name: "cant-write-header-value",
-				write_space: 24,
-				code: 200,
-				reason: "OK",
-				headers: &[
-					Header { name: "Foo", value: b"Bar" },
-				],
-				body_size: BodySize::Known(0),
-				ver_min: 0,
-				persistent: false,
-				result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
-				state: ServerState::AwaitingResponse,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.0 200 OK\r\nFoo: ",
-			},
-			Test {
-				name: "cant-write-header-eol",
-				write_space: 26,
-				code: 200,
-				reason: "OK",
-				headers: &[
-					Header { name: "Foo", value: b"Bar" },
-				],
-				body_size: BodySize::Known(0),
-				ver_min: 0,
-				persistent: false,
-				result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
-				state: ServerState::AwaitingResponse,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.0 200 OK\r\nFoo: Bar",
-			},
-			Test {
-				name: "cant-write-keep-alive",
-				write_space: 26,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::Known(0),
-				ver_min: 0,
-				persistent: true,
-				result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
-				state: ServerState::AwaitingResponse,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.0 200 OK\r\n",
-			},
-			Test {
-				name: "cant-write-close",
-				write_space: 26,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::Known(0),
-				ver_min: 1,
-				persistent: false,
-				result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
-				state: ServerState::AwaitingResponse,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.1 200 OK\r\n",
-			},
-			Test {
-				name: "cant-write-transfer-encoding",
-				write_space: 26,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::Unknown,
-				ver_min: 1,
-				persistent: false,
-				result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
-				state: ServerState::AwaitingResponse,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.1 200 OK\r\n",
-			},
-			Test {
-				name: "cant-write-content-length",
-				write_space: 26,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::Known(0),
-				ver_min: 0,
-				persistent: false,
-				result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
-				state: ServerState::AwaitingResponse,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.0 200 OK\r\n",
-			},
-			Test {
-				name: "cant-write-te-chunked",
-				write_space: 50,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::Unknown,
-				ver_min: 1,
-				persistent: true,
-				result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
-				state: ServerState::AwaitingResponse,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.1 200 OK\r\nConnection: Transfer-Encoding\r\n",
-			},
-			Test {
-				name: "cant-write-eol",
-				write_space: 18,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::Unknown,
-				ver_min: 0,
-				persistent: false,
-				result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
-				state: ServerState::AwaitingResponse,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.0 200 OK\r\n",
-			},
-			Test {
-				name: "exclude-headers",
-				write_space: 1024,
-				code: 200,
-				reason: "OK",
-				headers: &[
-					Header { name: "Connection", value: b"X" },
-					Header { name: "Foo", value: b"Bar" },
-					Header { name: "Content-Length", value: b"X" },
-					Header { name: "Transfer-Encoding", value: b"X" },
-				],
-				body_size: BodySize::Unknown,
-				ver_min: 0,
-				persistent: false,
-				result: Ok(()),
-				state: ServerState::SendingBody,
-				body_size_after: BodySize::Unknown,
-				chunked: false,
-				written: "HTTP/1.0 200 OK\r\nFoo: Bar\r\n\r\n",
-			},
-			Test {
-				name: "exclude-headers-101",
-				write_space: 1024,
-				code: 101,
-				reason: "Switching Protocols",
-				headers: &[
-					Header { name: "Connection", value: b"X" },
-					Header { name: "Foo", value: b"Bar" },
-					Header { name: "Content-Length", value: b"X" },
-					Header { name: "Transfer-Encoding", value: b"X" },
-				],
-				body_size: BodySize::NoBody,
-				ver_min: 0,
-				persistent: false,
-				result: Ok(()),
-				state: ServerState::SendingBody,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.0 101 Switching Protocols\r\nConnection: X\r\nFoo: Bar\r\n\r\n",
-			},
-			Test {
-				name: "1.0-no-body",
-				write_space: 1024,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::NoBody,
-				ver_min: 0,
-				persistent: false,
-				result: Ok(()),
-				state: ServerState::SendingBody,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.0 200 OK\r\n\r\n",
-			},
-			Test {
-				name: "1.0-len",
-				write_space: 1024,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::Known(42),
-				ver_min: 0,
-				persistent: false,
-				result: Ok(()),
-				state: ServerState::SendingBody,
-				body_size_after: BodySize::Known(42),
-				chunked: false,
-				written: "HTTP/1.0 200 OK\r\nContent-Length: 42\r\n\r\n",
-			},
-			Test {
-				name: "1.0-no-len",
-				write_space: 1024,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::Unknown,
-				ver_min: 0,
-				persistent: false,
-				result: Ok(()),
-				state: ServerState::SendingBody,
-				body_size_after: BodySize::Unknown,
-				chunked: false,
-				written: "HTTP/1.0 200 OK\r\n\r\n",
-			},
-			Test {
-				name: "1.1-no-body",
-				write_space: 1024,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::NoBody,
-				ver_min: 1,
-				persistent: true,
-				result: Ok(()),
-				state: ServerState::SendingBody,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.1 200 OK\r\n\r\n",
-			},
-			Test {
-				name: "1.1-len",
-				write_space: 1024,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::Known(42),
-				ver_min: 1,
-				persistent: true,
-				result: Ok(()),
-				state: ServerState::SendingBody,
-				body_size_after: BodySize::Known(42),
-				chunked: false,
-				written: "HTTP/1.1 200 OK\r\nContent-Length: 42\r\n\r\n",
-			},
-			Test {
-				name: "1.1-no-len",
-				write_space: 1024,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::Unknown,
-				ver_min: 1,
-				persistent: true,
-				result: Ok(()),
-				state: ServerState::SendingBody,
-				body_size_after: BodySize::Unknown,
-				chunked: true,
-				written: "HTTP/1.1 200 OK\r\nConnection: Transfer-Encoding\r\nTransfer-Encoding: chunked\r\n\r\n",
-			},
-			Test {
-				name: "1.0-persistent",
-				write_space: 1024,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::NoBody,
-				ver_min: 0,
-				persistent: true,
-				result: Ok(()),
-				state: ServerState::SendingBody,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.0 200 OK\r\nConnection: keep-alive\r\n\r\n",
-			},
-			Test {
-				name: "1.0-non-persistent",
-				write_space: 1024,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::NoBody,
-				ver_min: 0,
-				persistent: false,
-				result: Ok(()),
-				state: ServerState::SendingBody,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.0 200 OK\r\n\r\n",
-			},
-			Test {
-				name: "1.1-persistent",
-				write_space: 1024,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::NoBody,
-				ver_min: 1,
-				persistent: true,
-				result: Ok(()),
-				state: ServerState::SendingBody,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.1 200 OK\r\n\r\n",
-			},
-			Test {
-				name: "1.1-non-persistent",
-				write_space: 1024,
-				code: 200,
-				reason: "OK",
-				headers: &[],
-				body_size: BodySize::NoBody,
-				ver_min: 1,
-				persistent: false,
-				result: Ok(()),
-				state: ServerState::SendingBody,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n",
-			},
-			Test {
-				name: "force-no-body",
-				write_space: 1024,
-				code: 101,
-				reason: "Switching Protocols",
-				headers: &[],
-				body_size: BodySize::Known(42),
-				ver_min: 0,
-				persistent: false,
-				result: Ok(()),
-				state: ServerState::SendingBody,
-				body_size_after: BodySize::NoBody,
-				chunked: false,
-				written: "HTTP/1.0 101 Switching Protocols\r\n\r\n",
-			},
-		];
+            Test {
+                name: "cant-write-1.1",
+                write_space: 5,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::Known(0),
+                ver_min: 1,
+                persistent: false,
+                result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
+                state: ServerState::AwaitingResponse,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "",
+            },
+            Test {
+                name: "cant-write-1.0",
+                write_space: 5,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::Known(0),
+                ver_min: 0,
+                persistent: false,
+                result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
+                state: ServerState::AwaitingResponse,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "",
+            },
+            Test {
+                name: "cant-write-status-line",
+                write_space: 12,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::Known(0),
+                ver_min: 0,
+                persistent: false,
+                result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
+                state: ServerState::AwaitingResponse,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.0 200",
+            },
+            Test {
+                name: "cant-write-header-name",
+                write_space: 20,
+                code: 200,
+                reason: "OK",
+                headers: &[
+                    Header { name: "Foo", value: b"Bar" },
+                ],
+                body_size: BodySize::Known(0),
+                ver_min: 0,
+                persistent: false,
+                result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
+                state: ServerState::AwaitingResponse,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.0 200 OK\r\nFoo",
+            },
+            Test {
+                name: "cant-write-header-value",
+                write_space: 24,
+                code: 200,
+                reason: "OK",
+                headers: &[
+                    Header { name: "Foo", value: b"Bar" },
+                ],
+                body_size: BodySize::Known(0),
+                ver_min: 0,
+                persistent: false,
+                result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
+                state: ServerState::AwaitingResponse,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.0 200 OK\r\nFoo: ",
+            },
+            Test {
+                name: "cant-write-header-eol",
+                write_space: 26,
+                code: 200,
+                reason: "OK",
+                headers: &[
+                    Header { name: "Foo", value: b"Bar" },
+                ],
+                body_size: BodySize::Known(0),
+                ver_min: 0,
+                persistent: false,
+                result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
+                state: ServerState::AwaitingResponse,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.0 200 OK\r\nFoo: Bar",
+            },
+            Test {
+                name: "cant-write-keep-alive",
+                write_space: 26,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::Known(0),
+                ver_min: 0,
+                persistent: true,
+                result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
+                state: ServerState::AwaitingResponse,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.0 200 OK\r\n",
+            },
+            Test {
+                name: "cant-write-close",
+                write_space: 26,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::Known(0),
+                ver_min: 1,
+                persistent: false,
+                result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
+                state: ServerState::AwaitingResponse,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.1 200 OK\r\n",
+            },
+            Test {
+                name: "cant-write-transfer-encoding",
+                write_space: 26,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::Unknown,
+                ver_min: 1,
+                persistent: false,
+                result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
+                state: ServerState::AwaitingResponse,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.1 200 OK\r\n",
+            },
+            Test {
+                name: "cant-write-content-length",
+                write_space: 26,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::Known(0),
+                ver_min: 0,
+                persistent: false,
+                result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
+                state: ServerState::AwaitingResponse,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.0 200 OK\r\n",
+            },
+            Test {
+                name: "cant-write-te-chunked",
+                write_space: 50,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::Unknown,
+                ver_min: 1,
+                persistent: true,
+                result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
+                state: ServerState::AwaitingResponse,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.1 200 OK\r\nConnection: Transfer-Encoding\r\n",
+            },
+            Test {
+                name: "cant-write-eol",
+                write_space: 18,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::Unknown,
+                ver_min: 0,
+                persistent: false,
+                result: Err(ServerError::Io(io::Error::from(io::ErrorKind::WriteZero))),
+                state: ServerState::AwaitingResponse,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.0 200 OK\r\n",
+            },
+            Test {
+                name: "exclude-headers",
+                write_space: 1024,
+                code: 200,
+                reason: "OK",
+                headers: &[
+                    Header { name: "Connection", value: b"X" },
+                    Header { name: "Foo", value: b"Bar" },
+                    Header { name: "Content-Length", value: b"X" },
+                    Header { name: "Transfer-Encoding", value: b"X" },
+                ],
+                body_size: BodySize::Unknown,
+                ver_min: 0,
+                persistent: false,
+                result: Ok(()),
+                state: ServerState::SendingBody,
+                body_size_after: BodySize::Unknown,
+                chunked: false,
+                written: "HTTP/1.0 200 OK\r\nFoo: Bar\r\n\r\n",
+            },
+            Test {
+                name: "exclude-headers-101",
+                write_space: 1024,
+                code: 101,
+                reason: "Switching Protocols",
+                headers: &[
+                    Header { name: "Connection", value: b"X" },
+                    Header { name: "Foo", value: b"Bar" },
+                    Header { name: "Content-Length", value: b"X" },
+                    Header { name: "Transfer-Encoding", value: b"X" },
+                ],
+                body_size: BodySize::NoBody,
+                ver_min: 0,
+                persistent: false,
+                result: Ok(()),
+                state: ServerState::SendingBody,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.0 101 Switching Protocols\r\nConnection: X\r\nFoo: Bar\r\n\r\n",
+            },
+            Test {
+                name: "1.0-no-body",
+                write_space: 1024,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::NoBody,
+                ver_min: 0,
+                persistent: false,
+                result: Ok(()),
+                state: ServerState::SendingBody,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.0 200 OK\r\n\r\n",
+            },
+            Test {
+                name: "1.0-len",
+                write_space: 1024,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::Known(42),
+                ver_min: 0,
+                persistent: false,
+                result: Ok(()),
+                state: ServerState::SendingBody,
+                body_size_after: BodySize::Known(42),
+                chunked: false,
+                written: "HTTP/1.0 200 OK\r\nContent-Length: 42\r\n\r\n",
+            },
+            Test {
+                name: "1.0-no-len",
+                write_space: 1024,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::Unknown,
+                ver_min: 0,
+                persistent: false,
+                result: Ok(()),
+                state: ServerState::SendingBody,
+                body_size_after: BodySize::Unknown,
+                chunked: false,
+                written: "HTTP/1.0 200 OK\r\n\r\n",
+            },
+            Test {
+                name: "1.1-no-body",
+                write_space: 1024,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::NoBody,
+                ver_min: 1,
+                persistent: true,
+                result: Ok(()),
+                state: ServerState::SendingBody,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.1 200 OK\r\n\r\n",
+            },
+            Test {
+                name: "1.1-len",
+                write_space: 1024,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::Known(42),
+                ver_min: 1,
+                persistent: true,
+                result: Ok(()),
+                state: ServerState::SendingBody,
+                body_size_after: BodySize::Known(42),
+                chunked: false,
+                written: "HTTP/1.1 200 OK\r\nContent-Length: 42\r\n\r\n",
+            },
+            Test {
+                name: "1.1-no-len",
+                write_space: 1024,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::Unknown,
+                ver_min: 1,
+                persistent: true,
+                result: Ok(()),
+                state: ServerState::SendingBody,
+                body_size_after: BodySize::Unknown,
+                chunked: true,
+                written: "HTTP/1.1 200 OK\r\nConnection: Transfer-Encoding\r\nTransfer-Encoding: chunked\r\n\r\n",
+            },
+            Test {
+                name: "1.0-persistent",
+                write_space: 1024,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::NoBody,
+                ver_min: 0,
+                persistent: true,
+                result: Ok(()),
+                state: ServerState::SendingBody,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.0 200 OK\r\nConnection: keep-alive\r\n\r\n",
+            },
+            Test {
+                name: "1.0-non-persistent",
+                write_space: 1024,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::NoBody,
+                ver_min: 0,
+                persistent: false,
+                result: Ok(()),
+                state: ServerState::SendingBody,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.0 200 OK\r\n\r\n",
+            },
+            Test {
+                name: "1.1-persistent",
+                write_space: 1024,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::NoBody,
+                ver_min: 1,
+                persistent: true,
+                result: Ok(()),
+                state: ServerState::SendingBody,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.1 200 OK\r\n\r\n",
+            },
+            Test {
+                name: "1.1-non-persistent",
+                write_space: 1024,
+                code: 200,
+                reason: "OK",
+                headers: &[],
+                body_size: BodySize::NoBody,
+                ver_min: 1,
+                persistent: false,
+                result: Ok(()),
+                state: ServerState::SendingBody,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.1 200 OK\r\nConnection: close\r\n\r\n",
+            },
+            Test {
+                name: "force-no-body",
+                write_space: 1024,
+                code: 101,
+                reason: "Switching Protocols",
+                headers: &[],
+                body_size: BodySize::Known(42),
+                ver_min: 0,
+                persistent: false,
+                result: Ok(()),
+                state: ServerState::SendingBody,
+                body_size_after: BodySize::NoBody,
+                chunked: false,
+                written: "HTTP/1.0 101 Switching Protocols\r\n\r\n",
+            },
+        ];
 
         for test in tests.iter() {
             let mut p = ServerProtocol {
@@ -2604,6 +2779,35 @@ mod tests {
         );
 
         assert_eq!(str::from_utf8(&out).unwrap(), data);
+
+        let data = "GET /foo HTTP/1.1\r\nHost: example.com\r\n\r\n";
+
+        let mut p = ServerProtocol::new();
+        read_req(&mut p, data.as_bytes(), 2);
+
+        let mut resp = TestResponse::new();
+        resp.code = 200;
+        resp.reason = String::from("OK");
+        resp.headers = vec![(String::from("Content-Type"), b"text/plain".to_vec())];
+        resp.body = b"hello\n".to_vec();
+        resp.chunked = true;
+        resp.trailing_headers = vec![(String::from("Foo"), b"bar".to_vec())];
+
+        let out = write_resp(&mut p, resp, 2);
+
+        let data = concat!(
+            "HTTP/1.1 200 OK\r\n",
+            "Content-Type: text/plain\r\n",
+            "Connection: Transfer-Encoding\r\n",
+            "Transfer-Encoding: chunked\r\n",
+            "\r\n",
+            "6\r\nhello\n\r\n",
+            "0\r\n",
+            "Foo: bar\r\n",
+            "\r\n"
+        );
+
+        assert_eq!(str::from_utf8(&out).unwrap(), data);
     }
 
     #[test]
